from __future__ import annotations

from continual.python.sdk.exceptions import InvalidArgumentError


class ProjectEnvironmentIdentifer:
    """Utility class to handle project and environment names and identifers."""

    production_environment_ids = ["production", "prod"]
    default_environment_id = "production"

    def __init__(self, project_name_or_id="", environment_name_or_id=""):
        self.__project_name = ""
        self.__project_id = ""
        self.__environment_name = ""
        self.__environment_id = ""
        self.__environment_id_alias = ""
        self.__environment_is_default = False
        self.__environment_is_production = False
        self.__url_path = ""
        self.__parse(project_name_or_id, environment_name_or_id)

    def __parse(self, project_name_or_id: str, environment_name_or_id: str):
        project_part = ""
        environment_part = ""

        if project_name_or_id:
            if "@" in project_name_or_id:
                project_part = project_name_or_id.split("@")[0]
                environment_part = project_name_or_id.split("@")[1]
            else:
                project_part = project_name_or_id

        if environment_name_or_id:
            if "@" in environment_name_or_id:
                project_part = environment_name_or_id.split("@")[0]
                environment_part = environment_name_or_id.split("@")[1]
            else:
                environment_part = environment_name_or_id

        if not project_part:
            raise InvalidArgumentError(
                f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
            )
        elif project_part.startswith("projects/"):
            self.__project_id = project_part.split("projects/")[1]
            self.__project_name = project_part
        else:
            self.__project_id = project_part
            self.__project_name = f"projects/{self.__project_id}"

        if not environment_part:
            if self.__project_name:
                self.__environment_id = self.__project_name
                self.__environment_id_alias = self.default_environment_id
                self.__environment_name = (
                    f"{self.__project_name}@{self.__environment_id_alias}"
                )
                self.__environment_is_default = True
                self.__environment_is_production = True
            else:
                raise InvalidArgumentError(
                    f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
                )
        elif environment_part.startswith("projects/"):
            if environment_part == self.__project_name:
                self.__environment_id = environment_part
                self.__environment_id_alias = self.default_environment_id
                self.__environment_name = (
                    f"{environment_part}@{self.__environment_id_alias}"
                )
                self.__environment_is_default = True
                self.__environment_is_production = True
            else:
                raise InvalidArgumentError(
                    f"Default environment name does not match project name. Received: '{project_name_or_id}', '{environment_name_or_id}'"
                )
        elif environment_part in self.production_environment_ids:
            if self.__project_name:
                self.__environment_id = self.__project_name
                self.__environment_id_alias = self.default_environment_id
                self.__environment_name = (
                    f"{self.__project_name}@{self.__environment_id_alias}"
                )
                self.__environment_is_default = True
                self.__environment_is_production = True
            else:
                raise InvalidArgumentError(
                    f"A project name or ID is required. Received: '{project_name_or_id}', '{environment_name_or_id}'"
                )
        else:
            self.__environment_id = environment_part
            self.__environment_id_alias = environment_part
            self.__environment_is_default = False
            self.__environment_is_production = False
            self.__environment_name = f"{self.__project_name}@{self.__environment_id}"

        if self.__environment_is_default:
            self.__url_path = self.__project_name
        else:
            self.__url_path = self.__environment_name

    @property
    def project_name(self) -> str:
        """Project name `projects/[project_id]`."""
        return self.__project_name

    @property
    def project_id(self) -> str:
        """Project ID `[project_id]`."""
        return self.__project_id

    @property
    def environment_name(self) -> str:
        """Environment name `projects/[project_id]@[environment_id]`."""
        return self.__environment_name

    @property
    def environment_name_short(self) -> str:
        """Shortened environment name `[project_id]@[environment_id]`."""
        return self.__environment_name.replace("projects/", "")

    @property
    def environment_id(self) -> str:
        """Environment ID: `[environment_id]` when non-default, `[project_name]` when default."""
        return self.__environment_id

    @property
    def environment_id_alias(self) -> str:
        """Environment ID alias: `[environment_id]` when non-default, `[default_environment_id]` when default."""
        return self.__environment_id_alias

    @property
    def environment_is_default(self) -> str:
        """Whether the environment is the default environment."""
        return self.__environment_is_default

    @property
    def environment_is_production(self) -> str:
        """Whether the environment is the production environment."""
        return self.__environment_is_production

    @property
    def url_path(self) -> str:
        """The URL path for the project and environment: `projects/[project_id]` for projects and default environments, `projects/[project_id]@[environment_id]` for non-default environments."""
        return self.__url_path

    def __getitem__(self, key):
        return getattr(self, key)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/graphql/options.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/graphql/options.proto',
  package='continual.rpc.graphql',
  syntax='proto3',
  serialized_options=b'Z>github.com/continualiq/continual/continual/rpc/graphql;graphql',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#continual/rpc/graphql/options.proto\x12\x15\x63ontinual.rpc.graphql\x1a google/protobuf/descriptor.proto*8\n\x13GraphQLFieldOptions\x12\t\n\x05INPUT\x10\x00\x12\n\n\x06OUTPUT\x10\x01\x12\n\n\x06IGNORE\x10\x02:[\n\x07options\x12\x1d.google.protobuf.FieldOptions\x18\xd7\x08 \x01(\x0e\x32*.continual.rpc.graphql.GraphQLFieldOptions:1\n\treference\x12\x1d.google.protobuf.FieldOptions\x18\xd8\x08 \x01(\t:.\n\x05\x65\x64ges\x12\x1e.google.protobuf.MethodOptions\x18\xd9\x08 \x01(\t:0\n\x07parents\x12\x1e.google.protobuf.MethodOptions\x18\xda\x08 \x01(\t:4\n\x0bpassthrough\x12\x1e.google.protobuf.MethodOptions\x18\xdb\x08 \x01(\t:2\n\tedge_only\x12\x1e.google.protobuf.MethodOptions\x18\xdc\x08 \x01(\x08:/\n\x06ignore\x12\x1e.google.protobuf.MethodOptions\x18\xdd\x08 \x01(\x08\x42@Z>github.com/continualiq/continual/continual/rpc/graphql;graphqlb\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_descriptor__pb2.DESCRIPTOR,])

_GRAPHQLFIELDOPTIONS = _descriptor.EnumDescriptor(
  name='GraphQLFieldOptions',
  full_name='continual.rpc.graphql.GraphQLFieldOptions',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INPUT', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='OUTPUT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='IGNORE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=96,
  serialized_end=152,
)
_sym_db.RegisterEnumDescriptor(_GRAPHQLFIELDOPTIONS)

GraphQLFieldOptions = enum_type_wrapper.EnumTypeWrapper(_GRAPHQLFIELDOPTIONS)
INPUT = 0
OUTPUT = 1
IGNORE = 2

OPTIONS_FIELD_NUMBER = 1111
options = _descriptor.FieldDescriptor(
  name='options', full_name='continual.rpc.graphql.options', index=0,
  number=1111, type=14, cpp_type=8, label=1,
  has_default_value=False, default_value=0,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
REFERENCE_FIELD_NUMBER = 1112
reference = _descriptor.FieldDescriptor(
  name='reference', full_name='continual.rpc.graphql.reference', index=1,
  number=1112, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=b"".decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
EDGES_FIELD_NUMBER = 1113
edges = _descriptor.FieldDescriptor(
  name='edges', full_name='continual.rpc.graphql.edges', index=2,
  number=1113, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=b"".decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
PARENTS_FIELD_NUMBER = 1114
parents = _descriptor.FieldDescriptor(
  name='parents', full_name='continual.rpc.graphql.parents', index=3,
  number=1114, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=b"".decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
PASSTHROUGH_FIELD_NUMBER = 1115
passthrough = _descriptor.FieldDescriptor(
  name='passthrough', full_name='continual.rpc.graphql.passthrough', index=4,
  number=1115, type=9, cpp_type=9, label=1,
  has_default_value=False, default_value=b"".decode('utf-8'),
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
EDGE_ONLY_FIELD_NUMBER = 1116
edge_only = _descriptor.FieldDescriptor(
  name='edge_only', full_name='continual.rpc.graphql.edge_only', index=5,
  number=1116, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)
IGNORE_FIELD_NUMBER = 1117
ignore = _descriptor.FieldDescriptor(
  name='ignore', full_name='continual.rpc.graphql.ignore', index=6,
  number=1117, type=8, cpp_type=7, label=1,
  has_default_value=False, default_value=False,
  message_type=None, enum_type=None, containing_type=None,
  is_extension=True, extension_scope=None,
  serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key)

DESCRIPTOR.enum_types_by_name['GraphQLFieldOptions'] = _GRAPHQLFIELDOPTIONS
DESCRIPTOR.extensions_by_name['options'] = options
DESCRIPTOR.extensions_by_name['reference'] = reference
DESCRIPTOR.extensions_by_name['edges'] = edges
DESCRIPTOR.extensions_by_name['parents'] = parents
DESCRIPTOR.extensions_by_name['passthrough'] = passthrough
DESCRIPTOR.extensions_by_name['edge_only'] = edge_only
DESCRIPTOR.extensions_by_name['ignore'] = ignore
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

options.enum_type = _GRAPHQLFIELDOPTIONS
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(options)
google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(reference)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(edges)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(parents)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(passthrough)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(edge_only)
google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(ignore)

DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import typer
import pandas as pd

from continual.python.cli import utils
from continual import Client
from continual.python.cli.models import format_model_data
from continual.python.cli.batch_predictions import format_batch_prediction_data
from continual.python.cli.promotions import format_promotion_data
from continual.python.cli.experiments import format_experiments_data
from continual.python.cli.feature_sets import format_feature_set_data
from continual.python.cli.model_versions import format_model_version_data
from continual.python.cli.projects import format_project_data
from continual.python.cli.changes import format_changes_data

from rich.panel import Panel
from rich.padding import Padding
import rich

from typing import List
from enum import Enum

app = typer.Typer(help="Run reports.")


class GenericReportType(Enum):
    ALL = "ALL"
    ACTIVE = "ACTIVE"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"


class ModelReportType(Enum):
    ALL = "ALL"
    ACTIVE = "ACTIVE"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"
    HEALTHY = "HEALTHY"
    UNHEALTHY = "UNHEALTHY"
    OFFLINE = "OFFLINE"


class PromotionReportType(Enum):
    ALL = "ALL"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"


@app.command("models")
@utils.exit_on_error
def model_report(
    project: str = typer.Option(None, help="Project ID."),
    report_type: ModelReportType = typer.Argument(
        "ALL", help="Status to filter report on."
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters.", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report."),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    filters = list(filters_in)  # typer actually passes the List in as a tuple...
    if report_type == ModelReportType.ACTIVE:
        filters.append("latest_mv_state:CREATED")
        filters.append("latest_mv_state:PENDING")
        filters.append("latest_mv_state:PREPROCESSING")
        filters.append("latest_mv_state:TRAINING")
        filters.append("latest_mv_state:POSTPROCESSING")
    elif report_type == ModelReportType.FAILED:
        filters.append("latest_mv_state:CANCELLING")
        filters.append("latest_mv_state:FAILED")
        filters.append("latest_mv_state:CANCELLED")
        filters.append("latest_mv_state:STATE_UNSPECIFIED")
    elif report_type == ModelReportType.SUCCEEDED:
        filters.append("latest_mv_state:SUCCEEDED")
    elif report_type == ModelReportType.HEALTHY:
        filters.append("state:HEALTHY")
    elif report_type == ModelReportType.UNHEALTHY:
        filters.append("state:UNHEALTHY")
    elif report_type == ModelReportType.OFFLINE:
        filters.append("state:OFFLINE")

    data = []
    headers = []
    for m in c.models.list(n, filters=filters, all_projects=all_projects):
        (m_data, headers) = format_model_data(m, all_projects=all_projects)
        data.append(m_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Models found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        type_df = df.groupby(["Model Version Type"]).count()[["ID"]].reset_index()
        type_df.columns = ["Model Version Type", "Count"]
        type_table = utils.print_table(
            type_df,
            headers=type_df.columns.tolist(),
            print=False,
            title="Count by Type",
            style=style,
        )
        entity_df = df.groupby(["Entity"]).count()[["ID"]].reset_index()
        entity_df.columns = ["Entity", "Count"]
        entity_table = utils.print_table(
            entity_df,
            headers=entity_df.columns.tolist(),
            print=False,
            title="Count by Entity",
            style=style,
        )
        mv_count = df[["# Model Versions"]].sum()[0]
        pred_count = df[["# Batch Predictions"]].sum()[0]
        pro_count = df[["# Promotions"]].sum()[0]
        health_count = df[df["State"] == "HEALTHY"].count()[["ID"]][0]
        summary_panel = Padding(
            Panel(
                "%s Healthy Models out of %s\nTotal # of Model Versions: %s\nTotal # of Promotions: %s\nTotal # of Batch Prediction Jobs: %s"
                % (health_count, len(df), mv_count, pred_count, pro_count),
                style=style,
                title="Summary Stats",
                height=6,
            ),
            (1, 0, 0, 0),
        )

        # help text: the template below is equivalent to:
        # layout = Layout()
        # layout.split(
        #     Layout(main_table,name="Top", ratio=len(df)),
        #     Layout(name="Bottom", ratio=len(type_df)),
        # )
        # if all_projects:
        #     layout["Bottom"].split(
        #         Layout(project_table, name="Projects"),
        #         Layout(type_table,name="Types"),
        #         Layout(summary_panel,name="Summary Information"),
        #         direction= "horizontal"
        #     )
        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": max(len(type_df), len(entity_df)),
                "direction": "horizontal",
                "children": [
                    {"name": "Types", "data": type_table},
                    {"name": "Entities", "data": entity_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No models found for report type %s" % (report_type.value), fg="blue"
        )


@app.command("batch-predictions")
@utils.exit_on_error
def batch_prediction_report(
    project: str = typer.Option(None, help="Project to run report on"),
    report_type: GenericReportType = typer.Argument(
        "ALL", help="Status to filter report on"
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)
    if report_type == GenericReportType.ACTIVE:
        filters.append("state:CREATED")
        filters.append("state:PENDING")
        filters.append("state:PREPROCESSING")
        filters.append("state:PROCESSING")
        filters.append("state:POSTPROCESSING")
    elif report_type == GenericReportType.FAILED:
        filters.append("state:CANCELLING")
        filters.append("state:FAILED")
        filters.append("state:CANCELLED")
        filters.append("state:STATE_UNSPECIFIED")
    elif report_type == GenericReportType.SUCCEEDED:
        filters.append("state:SUCCEEDED")

    data = []
    headers = []
    for bpj in c.batch_prediction_jobs.list(
        n, filters=filters, all_projects=all_projects
    ):
        (bpj_data, headers) = format_batch_prediction_data(
            bpj, all_projects=all_projects
        )
        data.append(bpj_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Batch Prediction Jobs found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        model_df = df.groupby(["Model"]).count()[["ID"]].reset_index()
        model_df.columns = ["Model", "Count"]
        model_table = utils.print_table(
            model_df,
            headers=model_df.columns.tolist(),
            print=False,
            title="Count by Model",
            style=style,
        )
        # inc_df = df.groupby(["Incremental"]).count()[["ID"]].reset_index()
        # inc_df.columns = ["Incremental", "Count"]
        # inc_table = utils.print_table(
        #    inc_df, headers=inc_df.columns.tolist(), print=False, title="Count by Type")
        model_dur_df = df.groupby(["Model"]).mean()[["Duration (s)"]].reset_index()
        model_dur_df.columns = ["Model", "Average Duration"]
        model_dur_table = utils.print_table(
            model_dur_df,
            headers=model_dur_df.columns.tolist(),
            print=False,
            title="Avg Duration by Model",
            style=style,
        )
        pred_count = df[["# Predictions"]].sum()[0]
        duration = df[["Duration (s)"]].sum()[0]
        health_count = df[df["State"] == "SUCCEEDED"].count()[["ID"]][0]
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Successful Jobs out of %s\nTotal # of Predictions: %s\nAvg # of Predictions: %s\nTotal Time Spent Predicting: %s\nAvg Time Spent Predicting: %s"
                % (
                    health_count,
                    size,
                    pred_count,
                    pred_count / size,
                    duration,
                    duration / size,
                ),
                style=style,
                title="Summary Stats",
                height=7,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": max(len(model_df), len(model_dur_df)),
                "direction": "horizontal",
                "children": [
                    {"name": "Models", "data": model_table},
                    {"name": "Duration", "data": model_dur_table},
                    # {"name": "Type", "data": inc_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No batch predictions found for report type %s" % (report_type.value),
            fg="blue",
        )


@app.command("promotions")
@utils.exit_on_error
def promotion_report(
    project: str = typer.Option(None, help="Project to run report on"),
    report_type: PromotionReportType = typer.Argument(
        "ALL", help="Status to filter report on"
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)
    if report_type == GenericReportType.FAILED:
        filters.append("state:FAILED")
        filters.append("state:INVALID")
    elif report_type == GenericReportType.SUCCEEDED:
        filters.append("state:SUCCEEDED")

    data = []
    headers = []
    for pr in c.promotions.list(n, filters=filters, all_projects=all_projects):
        (pr_data, headers) = format_promotion_data(pr, all_projects=all_projects)
        data.append(pr_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Promotions found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        model_df = df.groupby(["Model"]).count()[["ID"]].reset_index()
        model_df.columns = ["Model", "Count"]
        model_table = utils.print_table(
            model_df,
            headers=model_df.columns.tolist(),
            print=False,
            title="Count by Model",
            style=style,
        )
        health_count = df[df["State"] == "SUCCEEDED"].count()[["ID"]][0]
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Successful Promotions out of %s" % (health_count, size),
                style=style,
                title="Summary Stats",
                height=3,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": len(model_df),
                "direction": "horizontal",
                "children": [
                    {"name": "Models", "data": model_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No promotions found for report type %s" % (report_type.value), fg="blue"
        )


@app.command("experiments")
@utils.exit_on_error
def experiment_report(
    project: str = typer.Option(None, help="Project to run report on"),
    report_type: GenericReportType = typer.Argument(
        "ALL", help="Status to filter report on"
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)
    if report_type == GenericReportType.ACTIVE:
        filters.append("state:CREATED")
        filters.append("state:PENDING")
        filters.append("state:PREPROCESSING")
        filters.append("state:TRAINING")
        filters.append("state:POSTPROCESSING")
    elif report_type == GenericReportType.FAILED:
        filters.append("state:CANCELLING")
        filters.append("state:FAILED")
        filters.append("state:CANCELLED")
        filters.append("state:STATE_UNSPECIFIED")
    elif report_type == GenericReportType.SUCCEEDED:
        filters.append("state:SUCCEEDED")

    data = []
    headers = []
    for e in c.experiments.list(n, filters=filters, all_projects=all_projects):
        (e_data, headers) = format_experiments_data(e, all_projects=all_projects)
        data.append(e_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Experiments found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        model_df = df.groupby(["Model"]).count()[["ID"]].reset_index()
        model_df.columns = ["Model", "Count"]
        model_table = utils.print_table(
            model_df,
            headers=model_df.columns.tolist(),
            print=False,
            title="Count by Model",
            style=style,
        )
        model_dur_df = df.groupby(["Model"]).mean()[["Duration (s)"]].reset_index()
        model_dur_df.columns = ["Model", "Average Duration"]
        model_dur_table = utils.print_table(
            model_dur_df,
            headers=model_dur_df.columns.tolist(),
            print=False,
            title="Avg Duration by Model",
            style=style,
        )
        # algo_dur_df = df.groupby(["Algorithm"]).mean()[
        #    ["Duration (s)"]].reset_index()
        # algo_dur_df.columns = ["Algorithm", "Average Duration"]
        # algo_dur_table = utils.print_table(
        #    algo_dur_df, headers=model_dur_df.columns.tolist(), print=False, title="Avg Duration by Algorithm")
        duration = df[["Duration (s)"]].sum()[0]
        health_count = df[df["State"] == "SUCCEEDED"].count()[["ID"]][0]
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Successful Experiments out of %s\nTotal Experiment Time: %s s\nAverage Experiment Time:%s s"
                % (health_count, size, duration, duration / size),
                style=style,
                title="Summary Stats",
                height=5,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": max(len(model_df), len(model_dur_df)),
                "direction": "horizontal",
                "children": [
                    {"name": "Models", "data": model_table},
                    {"name": "Models", "data": model_dur_table},
                    # {"name": "Models", "data": algo_dur_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No experiments found for report type %s" % (report_type.value), fg="blue"
        )


@app.command("feature-sets")
@utils.exit_on_error
def feature_set_report(
    project: str = typer.Option(None, help="Project to run report on"),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)

    data = []
    headers = []
    for fs in c.feature_sets.list(n, filters=filters, all_projects=all_projects):
        (fs_data, headers) = format_feature_set_data(fs, all_projects=all_projects)
        data.append(fs_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Feature Sets found (n=%s) %s" % (len(df), n, project_snippet),
            style=style,
        )
        entity_df = df.groupby(["Entity"]).sum()[["# Columns", "# Rows"]].reset_index()
        entity_df.columns = ["Entity", "# Columns", "# Rows"]
        entity_table = utils.print_table(
            entity_df,
            headers=entity_df.columns.tolist(),
            print=False,
            title="Stats by Entity",
            style=style,
        )
        feature_count = df[["# Columns"]].sum()[0]
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Total Feature Sets \nTotal Number of Entities: %s\nTotal \nTotal Number of Columns: %s"
                % (size, len(entity_df), feature_count),
                style=style,
                title="Summary Stats",
                height=5,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": len(entity_df),
                "direction": "horizontal",
                "children": [
                    {"name": "Entities", "data": entity_table},
                    {"name": "Stats", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho("No feature sets found", fg="blue")


# # ingestions
# @app.command("ingestion-report")
# @utils.exit_on_error
# def ingestion_report(
#     project: str = typer.Option(None, help="Project to run report on."),
#     report_type: GenericReportType = typer.Argument(
#         "ALL", help="Status to filter report on"
#     ),
#     all_projects: bool = typer.Option(
#         False, "--all", "-a", help="Filter on current project?"
#     ),
#     n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
#     filters_in: List[str] = typer.Option(
#         [], "--filter", "-f", help="List of filters", hidden=True
#     ),
# ):
#     c = Client(project=project)
#     if project is None:
#         project = utils.get_default_project()
#     project_snippet = " in ALL projects"
#     if not all_projects:
#         project_snippet = " in project %s" % project

#     # typer actually passes the List in as a tuple...
#     filters = list(filters_in)
#     if report_type == GenericReportType.ACTIVE:
#         filters.append("state:PENDING")
#         filters.append("state:RUNNING")
#     elif report_type == GenericReportType.FAILED:
#         filters.append("state:CANCELLING")
#         filters.append("state:FAILED")
#         filters.append("state:CANCELLED")
#         filters.append("state:STATE_UNSPECIFIED")
#     elif report_type == GenericReportType.SUCCEEDED:
#         filters.append("state:SUCCEEDED")

#     data = []
#     headers = []
#     for i in c.ingestions.list(n, filters=filters, all_projects=all_projects):
#         (i_data, headers) = format_ingestion_data(i, all_projects=all_projects)
#         data.append(i_data)

#     df = pd.DataFrame(data, columns=headers)
#     if len(df) > 0:
#         main_table = utils.print_table(
#             df,
#             headers=headers,
#             print=False,
#             title="%s Ingestions found for report type %s (n=%s) %s"
#             % (len(df), report_type.value, n, project_snippet),
#         )
#         fs_df = df.groupby(["Feature Set"]).count()[["ID"]].reset_index()
#         fs_df.columns = ["Feature Set", "Count"]
#         fs_table = utils.print_table(
#             fs_df,
#             headers=fs_df.columns.tolist(),
#             print=False,
#             title="Count by Feature Set",
#         )
#         ing_dur_df = df.groupby(["Feature Set"]).mean()[["Duration (s)"]].reset_index()
#         ing_dur_df.columns = ["Feature Set", "Average Duration"]
#         ing_dur_table = utils.print_table(
#             ing_dur_df,
#             headers=ing_dur_df.columns.tolist(),
#             print=False,
#             title="Avg Duration by Feature Set",
#         )
#         duration = df[["Duration (s)"]].sum()[0]
#         rows = df[["Ingested Rows"]].sum()[0]
#         health_count = df[df["State"] == "SUCCEEDED"].count()[["ID"]][0]
#         size = len(df)
#         summary_panel = Padding(
#             Panel(
#                 "%s Successful Ingestions out of %s\nTotal Ingestion Time: %s s\nAverage Ingestion Time:%s s\nTotal Rows Ingested: %s \nAverage Rows Ingested: %s"
#                 % (health_count, size, duration, duration / size, rows, rows / size),
#                 style=utils.continual_style_blue,
#                 title="Summary Stats",
#                 height=7,
#             ),
#             (1, 0, 0, 0),
#         )

#         template = [
#             {"name": "Top", "ratio": len(df), "data": main_table},
#             {
#                 "name": "Bottom",
#                 "ratio": max(len(fs_df), len(ing_dur_df)),
#                 "direction": "horizontal",
#                 "children": [
#                     {"name": "Models", "data": fs_table},
#                     {"name": "Models", "data": ing_dur_table},
#                     {"name": "Summary", "data": summary_panel},
#                 ],
#             },
#         ]
#         if all_projects:
#             project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
#             project_df.columns = ["Project", "Count"]
#             project_table = utils.print_table(
#                 project_df,
#                 headers=project_df.columns.tolist(),
#                 print=False,
#                 title="Count by Project",
#             )
#             template[1]["children"].insert(
#                 0, {"name": "Projects", "data": project_table}
#             )

#         layout = utils.build_dashboard(template, direction="vertical")
#         rich.print(layout)
#     else:
#         typer.secho(
#             "No ingestions found for report type %s" % (report_type.value), fg="blue"
#         )


# model versions
@app.command("model-versions")
@utils.exit_on_error
def model_version_report(
    project: str = typer.Option(None, help="Project to run report on"),
    report_type: GenericReportType = typer.Argument(
        "ALL", help="Status to filter report on"
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)
    if report_type == GenericReportType.ACTIVE:
        filters.append("state:CREATED")
        filters.append("state:PENDING")
        filters.append("state:PREPROCESSING")
        filters.append("state:TRAINING")
        filters.append("state:POSTPROCESSING")
    elif report_type == GenericReportType.FAILED:
        filters.append("state:CANCELLING")
        filters.append("state:FAILED")
        filters.append("state:CANCELLED")
        filters.append("state:STATE_UNSPECIFIED")
    elif report_type == GenericReportType.SUCCEEDED:
        filters.append("state:SUCCEEDED")

    data = []
    headers = []
    for mv in c.model_versions.list(n, filters=filters, all_projects=all_projects):
        (mv_data, headers) = format_model_version_data(mv, all_projects=all_projects)
        data.append(mv_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Model Versions found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        model_df = df.groupby(["Model"]).count()[["ID"]].reset_index()
        model_df.columns = ["Model", "Count"]
        model_table = utils.print_table(
            model_df,
            headers=model_df.columns.tolist(),
            print=False,
            title="Count by Model",
            style=style,
        )
        metric_df = (
            df.groupby(["Model", "Metric"])[["Metric Val"]]
            .agg(["min", "max"])
            .reset_index()
        )
        metric_df.columns = ["Model", "Metric", "Min", "Max"]
        metric_table = utils.print_table(
            metric_df,
            headers=metric_df.columns.tolist(),
            print=False,
            title="Min/Max Metric by Model",
            style=style,
        )
        promoted = len(df[df["Promoted"] == True])
        health_count = len(df[df["State"] == "SUCCEEDED"])
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Successful Model Versions out of %s\nTotal Promoted Model Versions: %s"
                % (health_count, size, promoted),
                style=style,
                title="Summary Stats",
                height=4,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": max(len(metric_df), len(model_df)),
                "direction": "horizontal",
                "children": [
                    {"name": "Models", "data": model_table},
                    {"name": "Metrics", "data": metric_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No model versions found for report type %s" % (report_type.value),
            fg="blue",
        )


# proj
@app.command("projects")
@utils.exit_on_error
def project_report(
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client()
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)

    data = []
    headers = []
    for p in c.projects.list(n, filters=filters):
        (p_data, headers) = format_project_data(p)
        data.append(p_data)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Projects Found (n=%s)" % (len(df), n),
            style=style,
        )
        org_df = df.groupby(["Organization"]).count()[["ID"]].reset_index()
        org_df.columns = ["Organizaiton", "Count"]
        org_table = utils.print_table(
            org_df,
            headers=org_df.columns.tolist(),
            print=False,
            title="Count by Organization",
            style=style,
        )
        fstore_df = df.groupby(["FStore Type"]).count()[["ID"]].reset_index()
        fstore_df.columns = ["Feature Store Type", "Count"]
        fstore_table = utils.print_table(
            fstore_df,
            headers=fstore_df.columns.tolist(),
            print=False,
            title="Count by Feature Store Type",
            style=style,
        )
        fs_count = df[["# Feature Sets"]].sum()[0]
        feature_count = df[["# Features"]].sum()[0]
        model_count = df[["# Models"]].sum()[0]
        mv_count = df[["# Model Versions"]].sum()[0]
        prediction_count = df[["# Predictions"]].sum()[0]
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Projects Total\nTotal Feature Sets: %s\nTotal Features: %s\nTotal Models: %s\nTotal Model Versions: %s \nTotal Predictions: %s"
                % (
                    size,
                    fs_count,
                    feature_count,
                    model_count,
                    mv_count,
                    prediction_count,
                ),
                style=style,
                title="Summary Stats",
                height=8,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": max(len(fstore_df), len(org_df)),
                "direction": "horizontal",
                "children": [
                    {"name": "Models", "data": org_table},
                    {"name": "Metrics", "data": fstore_table},
                    {"name": "Summary", "data": summary_panel},
                ],
            },
        ]

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho("No projects found", fg="blue")


# changes


@app.command("changes")
@utils.exit_on_error
def changes_report(
    project: str = typer.Option(None, help="Project to run report on"),
    report_type: GenericReportType = typer.Argument(
        "ALL", help="Status to filter report on"
    ),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects"),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show"),
    filters_in: List[str] = typer.Option(
        [], "--filter", "-f", help="List of filters", hidden=True
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for report"),
):
    c = Client(project=project)
    if project is None:
        project = utils.get_default_project()
    project_snippet = " in ALL projects"
    if not all_projects:
        project_snippet = " in project %s" % project
    if style is None:
        style = c.config.style
    style = utils.get_style(style)

    # typer actually passes the List in as a tuple...
    filters = list(filters_in)
    if report_type == GenericReportType.ACTIVE:
        filters.append("state:PENDING")
        filters.append("state:EXECUTING")
    elif report_type == GenericReportType.FAILED:
        filters.append("state:CANCELLING")
        filters.append("state:FAILED")
        filters.append("state:CANCELLED")
        filters.append("state:STATE_UNSPECIFIED")
    elif report_type == GenericReportType.SUCCEEDED:
        filters.append("state:SUCCEEDED")

    data = []
    headers = []
    pushes = c.changes.list(n, filters=filters, all_projects=all_projects)
    (data, headers) = format_changes_data(pushes, project, n, filters, all_projects)

    df = pd.DataFrame(data, columns=headers)
    if len(df) > 0:
        main_table = utils.print_table(
            df,
            headers=headers,
            print=False,
            title="%s Changes found for report type %s (n=%s) %s"
            % (len(df), report_type.value, n, project_snippet),
            style=style,
        )
        health_count = len(df[df["State"] == "SUCCEEDED"])
        size = len(df)
        summary_panel = Padding(
            Panel(
                "%s Successful Changes out of %s" % (health_count, size),
                style=style,
                title="Summary Stats",
                height=3,
            ),
            (1, 0, 0, 0),
        )

        template = [
            {"name": "Top", "ratio": len(df), "data": main_table},
            {
                "name": "Bottom",
                "ratio": 1,
                "direction": "horizontal",
                "children": [{"name": "Summary", "data": summary_panel}],
            },
        ]
        if all_projects:
            project_df = df.groupby(["Project"]).count()[["ID"]].reset_index()
            project_df.columns = ["Project", "Count"]
            project_table = utils.print_table(
                project_df,
                headers=project_df.columns.tolist(),
                print=False,
                title="Count by Project",
                style=style,
            )
            template[1]["children"].insert(
                0, {"name": "Projects", "data": project_table}
            )

        layout = utils.build_dashboard(template, direction="vertical")
        rich.print(layout)
    else:
        typer.secho(
            "No changes found for report type %s" % (report_type.value), fg="blue"
        )


# users
## TODO
# @app.command("user-report")
# @utils.exit_on_error
# def user_report(

# org
## TODO
# @app.command("organization-report")
# @utils.exit_on_error
# def organization_report(

# events
## TODO
# @app.command("event-report")
# @utils.exit_on_error
# def event_report(


# feature store
## TODO
# @app.command("feature-store-report")
# @utils.exit_on_error
# def feature_store_report(



import base64
import sys
import logging
from typing import List, Tuple
from google.protobuf.timestamp_pb2 import Timestamp

import pandas as pd

from continual.rpc.management.v1 import types

base64_encode_prefix = "continual_"


def get_parent(name: str) -> str:
    parts = name.split("/")
    if len(parts) < 2:
        return ""
    return "/".join(parts[:-2])


def parse_project_name(project_name):
    # projects/<project>
    items = project_name.split("/")
    if len(items) != 2 or items[0] != "projects" or not items[1]:
        raise Exception("Error parsing project name [{}]".format(project_name))
    return items[1]


def parse_model_version_name(model_version_name):
    # projects/<project>/models/<model>/versions/<model_version>
    items = model_version_name.split("/")
    if (
        len(items) != 6
        or items[0] != "projects"
        or items[2] != "models"
        or items[4] != "versions"
        or not items[1]
        or not items[3]
        or not items[5]
    ):
        raise Exception(
            "Error parsing model version name [{}]".format(model_version_name)
        )

    return (items[1], items[3], items[5])


def parse_experiment_name(experiment_name):
    # projects/<project>/models/<model>/versions/<model_version>/experiments/<experiment>
    items = experiment_name.split("/")
    if (
        len(items) != 8
        or items[0] != "projects"
        or items[2] != "models"
        or items[4] != "versions"
        or items[6] != "experiments"
        or not items[1]
        or not items[3]
        or not items[5]
        or not items[7]
    ):
        raise Exception("Error parsing experiment name [{}]".format(experiment_name))

    return (items[1], items[3], items[5], items[7])


def validate_experiment_name(experiment_name):
    # projects/<project>/versions/<model_version>/experiments/<experiment>
    items = experiment_name.split("/")
    if (
        len(items) != 8
        or items[0] != "projects"
        or items[2] != "models"
        or items[4] != "versions"
        or items[6] != "experiments"
        or not items[1]
        or not items[3]
        or not items[5]
        or not items[7]
    ):
        raise Exception("Error parsing experiment name [{}]".format(experiment_name))
    return


def is_integer_value(val):
    try:
        int(val)
    except ValueError:
        return False
    return True


def get_experiment_id_from_name(experiment_name):
    validate_experiment_name(experiment_name)

    # The last field is the experiment ID
    return experiment_name.split("/")[-1]


def validate_project_name(project_name):
    # projects/<project_id>
    items = project_name.split("/")
    if len(items) != 2 or items[0] != "projects" or not items[1]:
        raise Exception("Error parsing project name [{}]".format(project_name))
    return


def validate_model_version_name(model_version_name):
    parse_model_version_name(model_version_name)
    return


def get_project_name(project_id):
    # projects/<project_id>
    if not project_id:
        raise Exception("project ID should not be an empty string")
    return "projects/{}".format(project_id)


def parse_model_name(model_name: str) -> Tuple[str, str]:
    # projects/:project_id/models/:model_id
    items = model_name.split("/")
    if (
        len(items) != 4
        or items[0] != "projects"
        or items[2] != "models"
        or not items[1]
        or not items[3]
    ):
        raise Exception("Error parsing model name [{}]".format(model_name))
    return items[1], items[3]


def parse_batch_prediction_job_name(batch_prediction_job_name):
    # projects/<project>/models/<model>/batchPredictions/<batch_prediction_job_id>
    items = batch_prediction_job_name.split("/")
    if (
        len(items) != 6
        or items[0] != "projects"
        or items[2] != "models"
        or items[4] != "batchPredictions"
        or not items[1]
        or not items[3]
        or not items[5]
    ):
        raise Exception(
            "Error parsing batch prediction job name [{}]".format(
                batch_prediction_job_name
            )
        )

    return (items[1], items[3], items[5])


def validate_model_name(model_name):
    parse_model_name(model_name)
    return


def get_model_version_name(model_name, model_version_id):
    validate_model_name(model_name)
    return "{}/versions/{}".format(model_name, model_version_id)


def get_featureset_name(project_name, feature_set_id):
    return "{}/featureSets/{}".format(project_name, feature_set_id)


def get_experiment_name(model_version_name, experiment_id):
    return "{}/experiments/{}".format(model_version_name, experiment_id)


def get_featureset_from_modelversion(model_version_name):
    project_id, model_id, _ = parse_model_version_name(model_version_name)
    entity = model_id.split(":")[0]
    fs_name = get_featureset_name(get_project_name(project_id), entity)
    return fs_name


def _get_float_replacement(val, name, mapper):
    if mapper.get(name, None) is not None:
        return mapper[name]
    return val


float_dict = {"inf": sys.float_info.max, "-inf": sys.float_info.min}


def renormalize_floats(obj, mapper):
    if isinstance(obj, list):
        out = []
        for val in obj:
            out.append(renormalize_floats(val, mapper))
        return out
    if isinstance(obj, dict):
        for key, val in obj.items():
            obj[key] = renormalize_floats(val, mapper)
        return obj
    if isinstance(obj, float):
        if obj == float("inf"):
            obj = _get_float_replacement(obj, "inf", mapper)
        elif obj == float("-inf"):
            obj = _get_float_replacement(obj, "-inf", mapper)
        elif obj == float("nan"):
            obj = _get_float_replacement(obj, "nan", mapper)
        return obj
    return obj


def parse_feature_set_name(name):
    """
    Splits name string such as 'projects/<project>/featureSets/<featureset>@<version>' to a dict like:
     {
       'projects' : <project>,
       'featureSets' : <featureset>s
     }
    """
    retmap = {}
    splits = name.split("/")
    i = 0
    while i < len(splits):
        retmap[splits[i]] = None
        if i + 1 < len(splits):
            retmap[splits[i]] = splits[i + 1]
        i = i + 2

    return retmap


def get_current_time():
    ts = Timestamp()
    ts.GetCurrentTime()
    return ts


def get_model_name(project_name, model_id, validate_fields=True):
    # projects/project_id/models/model_id
    if validate_fields:
        validate_project_name(project_name)
    return "{}/models/{}".format(project_name, model_id)


def encode_featurename_base64(name):
    continual_name = bytes(base64_encode_prefix + name, encoding="utf-8")
    out_bytes = base64.urlsafe_b64encode(continual_name)
    out = str(out_bytes, encoding="utf-8")
    return out.rstrip("=")


def encode_featurename_list(names_lst):
    out = []
    for name in names_lst:
        out.append(encode_featurename_base64(name))
    return out


def decode_featurename_base64(base64val):
    if not isinstance(base64val, str):
        return base64val
    base64str = base64val
    n = len(base64str)
    pad_str = ((4 - n % 4) % 4) * "="
    padded_base64str = base64str + pad_str
    try:
        out_bytes = base64.urlsafe_b64decode(bytes(padded_base64str, encoding="utf-8"))
        out = str(out_bytes, encoding="utf-8")
        if not out.startswith(base64_encode_prefix):
            return base64str
        return out[len(base64_encode_prefix) :]
    except Exception:
        return base64str
    return base64str


def decode_base64_obj(base64obj):
    if isinstance(base64obj, list):
        out_list = []
        for item in base64obj:
            out_list.append(decode_base64_obj(item))
        return out_list
    if isinstance(base64obj, dict):
        out_dict = {}
        for key, val in base64obj.items():
            key = decode_featurename_base64(key)
            out_dict[key] = decode_base64_obj(val)
        return out_dict
    if isinstance(base64obj, str):
        return decode_featurename_base64(base64obj)
    return base64obj


# This is an in-place encode of column names in df
def encode_df_column_names(df):
    columns = [encode_featurename_base64(column_name) for column_name in df.columns]
    df.columns = columns
    return


# This is an in-place decode of column names in df
def decode_df_column_names(df):
    columns = [decode_featurename_base64(column_name) for column_name in df.columns]
    df.columns = columns
    return


def split_name(name):
    splits = name.split("/")
    i = 0
    name_map = {}
    while i < len(splits):
        name_map[splits[i]] = splits[i + 1]
        i = i + 2
    return splits[len(splits) - 1], name_map


def dot_get(obj, path, raise_error=True):
    ret = None
    paths, current = path.split("."), obj
    for p in paths:
        if p not in current:
            if raise_error:
                raise KeyError("Could not find key in object")

            break
        elif p in current:
            current = current[p]
            ret = current

    return ret


def read_csv(file_path, chunksize=None):
    encoding = "utf-8"
    try:
        with open(file_path, "r") as f:
            logging.info("File ({}) has encoding: {}".format(file_path, f.encoding))
            encoding = f.encoding
    except Exception as e:
        logging.info(
            "Failed to open file locally: file_path={}, error: {}".format(
                file_path, str(e)
            )
        )

    try:
        return pd.read_csv(
            file_path, encoding=encoding, low_memory=False, chunksize=chunksize
        )
    except UnicodeDecodeError:
        logging.info(
            "Ran into unicode decode error with encoding ({}) — retrying with latin-1".format(
                encoding
            )
        )
        return pd.read_csv(
            file_path, encoding="latin-1", low_memory=False, chunksize=chunksize
        )


def generate_field_mask_paths(obj: dict) -> List[str]:
    paths = []

    q = list(obj.keys())
    while q:
        path = q.pop()

        value = dot_get(obj, path, raise_error=False)
        if value is not None and type(value) is dict:
            if len(value.keys()) == 0:
                paths.append(path)
            else:
                for k in value.keys():
                    q.append(f"{path}.{k}")
        else:
            paths.append(path)

    return paths


def generate_datastore_field_paths(data_store: types.DataStore) -> List[str]:
    paths = []
    if type(data_store) is not dict:
        data_store = data_store.to_dict()
    for path in generate_field_mask_paths(data_store):
        qualified_path = f"data_store.config.{path}"
        paths.append(qualified_path)
        if qualified_path.endswith(".username"):
            paths.append(qualified_path.replace(".username", ".user"))
    return paths



import typer
import yamale
import os
import yaml

from continual.python.cli import utils
from continual import Client

from pathlib import Path
from rich.console import Console
from typing import List

app = typer.Typer(help="Manage feature set.")


def format_feature_set_data(fs, zipped=False, all_projects=False):
    create_time = fs.create_time.replace(microsecond=0)
    update_time = fs.update_time.replace(microsecond=0)
    if zipped:
        data = [
            fs.id,
            fs.name,
            fs.schema.entity,
            len(fs.schema.columns),
            create_time,
            update_time,
            fs.schema.description[:50],
        ]
        headers = [
            "ID",
            "Name",
            "Entity",
            "Columns",
            "Created",
            "Last Updated",
            "Description",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            fs.id,
            fs.schema.entity,
            len(fs.schema.columns),
            create_time,
            update_time,
            fs.schema.description[:50],
        ]
        headers = [
            "ID",
            "Entity",
            "Columns",
            "Created",
            "Last Updated",
            "Description",
        ]
        if all_projects:
            data.insert(0, fs.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option(
        [],
        "--filter",
        "-f",
        help="List of filters",
    ),
    all_projects: bool = typer.Option(
        False,
        "--all",
        "-a",
        help="Show all projects",
    ),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List feature sets."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if all_projects:
        project_snippet = "all accessible projects"
    if len(filters) > 0:
        filter_snippet = " with filter %s" % str(filters) + filter_snippet
    data = []
    headers = []
    for fs in c.feature_sets.list(n, filters=filters, all_projects=all_projects):
        (fs_data, headers) = format_feature_set_data(fs, all_projects=all_projects)
        data.append(fs_data)
    typer.secho(
        "\nFound %s feature sets in %s%s:"
        % (len(data), project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    feature_set: str = typer.Argument(..., help="Feature set ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    json: bool = typer.Option(
        False, "--json", help="Display full JSON representation."
    ),
):
    """Get feature set details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    fs = c.feature_sets.get(feature_set)
    console = Console()
    if json:
        console.print(fs.to_dict())
    else:
        data = format_feature_set_data(fs, zipped=True)
        typer.secho("\nRetrieving feature set %s: \n" % (feature_set), fg="blue")
        utils.print_info(data)
        typer.secho("\nFeature set schema:\n", fg="blue")
        console.print(fs.schema.to_dict())


# dev-2222: removing until we have a better way of downloading data
# @app.command("get-data")
# @utils.exit_on_error
# def get_data(
#     feature_set: str = typer.Argument(..., help="Feature set ID."),
#     project: str = typer.Option(None, help="Project ID."),
#     environment: str = typer.Option(None, "--env", help="Environment ID."),
#     path: Path = typer.Option(..., help="Path to save CSV file."),
#     num_rows: int = typer.Option(None, "--n", help="Number of rows to fetch."),
# ):
#     """Get feature set data."""
#     project, environment = utils.get_project_and_environment(project, environment)
#     c = Client(project=utils.get_environment_name(project, environment))
#     fs = c.feature_sets.get(feature_set)
#     path = str(path)

#     try:
#         typer.secho("Starting download of feature set %s." % feature_set, fg="blue")
#         (df, _) = fs.get_data(page_size=num_rows)
#         if path[-4:] != ".csv":
#             path = "%s/%s.csv" % (path, feature_set)
#         if df is not None:
#             df.to_csv(path, index=False)
#             typer.secho(
#                 "Successfully saved data for featureset %s to %s"
#                 % (feature_set, str(path)),
#                 fg="green",
#             )
#         else:
#             typer.secho(
#                 "Failed to download data for feature set %s" % feature_set, fg="red"
#             )
#     except Exception as e:
#         typer.secho(
#             "Failed to get data for feature set %s in project %s, environment %s:%s"
#             % (feature_set, project, environment, str(e)),
#             fg="red",
#         )


@app.command("delete")
@utils.exit_on_error
def delete(
    feature_set: str = typer.Argument(..., help="Feature set ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    force: bool = typer.Option(
        False, "--force", help="Force deletion. Skips confirmation."
    ),
):
    """Delete feature set."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))
    if not force:
        force = typer.confirm(
            "Are you sure you want to delete feature set %s in project %s, environment %s?"
            % (feature_set, project, environment)
        )
    if force:
        fs = c.feature_sets.get(feature_set)
        fs.delete()
        typer.secho("Successfully deleted feature set %s" % fs.name, fg="green")
    else:
        typer.secho("Deletion Cancelled!", fg="blue")


@app.command("validate")
@utils.exit_on_error
def validate_yaml(
    path: Path = typer.Argument(..., help="Full path of YAML file."),
):
    """Validate feature set YAML file."""
    path = str(path)
    try:
        with open(path, "r") as file:
            text = yaml.load(file, Loader=yaml.FullLoader)
            yaml_type = text.get("type", "").lower()

        if yaml_type == "featureset" or yaml_type == "model":
            schema = yamale.make_schema(
                "%s/%s"
                % (
                    os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
                    "extras/%s_reference_schema.yaml" % yaml_type,
                )
            )
            data = yamale.make_data(path)
            yamale.validate(schema, data)
            typer.secho("Schema %s validated. No Errors found!" % path, fg="green")
            return True
        else:
            typer.secho(
                "Given schema does not have a valid type field. Cannot validate. type: %s"
                % yaml_type,
                fg="red",
            )
            return False
    except Exception as e:
        typer.secho("Schema validation for %s failed: \n%s" % (path, str(e)), fg="red")
        return False


@app.command("export")
@utils.exit_on_error
def export_yaml(
    feature_set: str = typer.Argument(..., help="Feature set ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    path: str = typer.Option(
        "print",
        help="Path to save YAML (local). Default = print to console.",
        show_default=False,
    ),
):
    """Export YAML file for feature set."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    fs = c.feature_sets.get(feature_set)
    fs_text = yaml.load(fs.schema_text, Loader=yaml.FullLoader)
    yaml_text = utils.format_yaml(fs_text)
    if path == "print":
        typer.secho(
            "Successfully extracted YAML file from feature set %s." % feature_set,
            fg="green",
        )
        print(yaml_text)
    else:
        utils.save_yamls([(feature_set, yaml_text)], path)
        typer.secho(
            "Successfully extracted YAML file from feature set %s. Location is %s."
            % (feature_set, path),
            fg="green",
        )


@app.command("diff")
@utils.exit_on_error
def diff(
    right_yaml: str = typer.Option(None, "--ryaml", help="YAML location."),
    right_project: str = typer.Option(None, "--rproj", help="Right project."),
    right_environment: str = typer.Option(None, "--renv", help="Environment ID."),
    right_feature_set: str = typer.Option(None, "--rfs", help="Right feature set."),
    left_yaml: str = typer.Option(None, "--lyaml", help="YAML location."),
    left_project: str = typer.Option(None, "--lproj", help="Left project."),
    left_environment: str = typer.Option(None, "--lenv", help="Environment ID."),
    left_feature_set: str = typer.Option(None, "--lfs", help="Left feature set"),
    models: bool = typer.Option(
        False, hidden=True, help="Diff models instead of feature sets"
    ),
):
    """Diff feature sets.

    Diff feature sets can either provide a local YAML file
    or a feature set id already contained in continual. To use YAML files,
    use the *_yaml options, otherwise specify the continual resources
    via *_project and *_feature_set
    """
    if not ((right_yaml or right_feature_set) and (left_yaml or left_feature_set)):
        typer.secho(
            "Error: You must either provide 2 YAML files or 2 feature-sets. See continual feature-sets diff --help."
        )
    else:
        c = Client()
        (right, left, diff_text) = c.feature_sets.diff(
            right_yaml,
            right_project,
            right_environment,
            right_feature_set,
            left_yaml,
            left_project,
            left_environment,
            left_feature_set,
            models,
        )
        typer.secho(
            "Diffing \n\tleft = %s with \n\tright = %s" % (left, right), fg="blue"
        )
        if len(diff_text) > 0:
            typer.secho(diff_text)
        else:
            typer.secho("No changes found.", fg="blue")



from typing import Any, Dict, Set


class bidict(dict):
    def __init__(self, mapping: Dict[Any, Any]) -> None:
        self._dict: Dict[Any, Any] = {}
        self.inverse: Dict[Any, Set[Any]] = {}
        for key, value in mapping.items():
            self[key] = value
        return

    def __setitem__(self, key: Any, value: Any) -> None:
        self._dict[key] = value
        self.inverse.setdefault(value, set()).add(key)
        return

    def __getitem__(self, key: Any) -> Any:
        return self._dict[key]

    def __delitem__(self, key: Any) -> None:
        self.inverse[self._dict[key]].remove(key)
        del self._dict[key]
        return

    def __repr__(self) -> str:
        return repr(self._dict)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['continual',
 'continual.python',
 'continual.python.cli',
 'continual.python.cli.tests',
 'continual.python.common',
 'continual.python.sdk',
 'continual.python.utils',
 'continual.rpc',
 'continual.rpc.graphql',
 'continual.rpc.management',
 'continual.rpc.management.v1',
 'continual.rpc.rpc']

package_data = \
{'': ['*'],
 'continual.python': ['examples/bank_marketing/*',
                      'examples/kickstarter/*',
                      'extras/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'cron-descriptor>=1.2.24,<2.0.0',
 'fsspec>=0.8.5,<0.9.0',
 'gitpython>=3.1.7,<4.0.0',
 'grpcio>=1.27.1,<2.0.0',
 'grpcio_status>=1.31.0,<2.0.0',
 'halo>=0.0.30,<0.0.31',
 'humanize>=2.5.0,<3.0.0',
 'pandas-gbq>=0.14.1,<0.15.0',
 'pandas>=1.0.1,<2.0.0',
 'pytz>=2020.5,<2021.0',
 'pyyaml>=5.3,<6.0',
 'requests==2.23.0',
 'rich>=9.13.0,<10.0.0',
 'tabulate>=0.8.6,<0.9.0',
 'tqdm>=4.54.1,<5.0.0',
 'typer>=0.3.2,<0.4.0',
 'yamale>=3.0.4,<4.0.0']

entry_points = \
{'console_scripts': ['continual = continual.python.cli.cli:cli']}

setup_kwargs = {
    'name': 'continual',
    'version': '0.5.27',
    'description': 'Operational AI for the Modern Data Stack',
    'long_description': '# Python CLI and SDK for Continual\n\nContinual is an operational AI for the modern data stack. Learn more at\nhttps://continual.ai.\n\n## Getting Started\n\nTo install the Continual CLI and SDK run:\n\n```\npip3 install continual\n```\n',
    'author': 'Continual',
    'author_email': 'support@continual.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import typing

# Required to avoid circular dependency issues:
# https://www.stefaanlippens.net/circular-imports-type-hints-python.html
if typing.TYPE_CHECKING:
    from continual.python.sdk.client import Client


class Resource:
    """Base resource class."""

    name: str
    name_pattern: str

    @classmethod
    def make_name(cls, parent: str, id: str) -> str:
        """Generates a resource name from parent and id.

        Adds wildcard for any missing parent elements.

        Arguments:
            parent: Parent resource name.
            id: Resource id.

        Returns:
            Resource name.
        """
        pattern_parts = cls.name_pattern.split("/")
        parent = parent or ""
        parent_parts = parent.split("/")
        out = []
        for i in range(len(pattern_parts)):
            if (i + 1) % 2 == 0:
                if i < len(parent_parts):
                    out.append(parent_parts[i])
                else:
                    out.append("-")
            else:
                out.append(pattern_parts[i])
        out[-1] = id
        return "/".join(out)

    @property
    def id(self) -> str:
        """ID of resource."""
        return self.name.split("/").pop()

    @property
    def parent(self) -> str:
        """Parent resource name."""
        return "/".join(self.name.split("/")[0:-2])

    @property
    def client(self) -> Client:
        """Continual Client object."""
        return self._client



from __future__ import annotations
from typing import List, Optional

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager


class EventManager(Manager):
    """Manages event resources."""

    # name_pattern: str = "projects/{project}/models/{model}"

    def get(self, id: str) -> Event:
        """Get Event.

        Arguments:
            id: Event id.

        Returns
            An Event.
        """
        req = management_pb2.GetEventRequest(name=self.name(id))
        resp = self.client._management.GetEvent(req)
        return Event.from_proto(resp, client=self.client)

    def list(
        self,
        filters: List[str] = None,
        page_size: Optional[int] = None,
        parent: str = None,
        all_projects=False,
    ) -> List[Event]:
        """List Events.

        Arguments:
            filter: List of filters to apply to events. Can be:
                - event type (i.e. event:featureSet:CREATED)
                - severity (i.e. severity:CRITICAL)
                - subject (i.e. subject:userAccount/<id>)

            page_size: Number of items to return.

            parent: Parent resource to filter events by

        Returns:
            A list of events.
        """
        if parent is None:
            parent = self.parent
        req = management_pb2.ListEventsRequest(
            parent=parent,
            filters=filters,
            page_size=page_size,
            all_projects=all_projects,
        )
        resp = self.client._management.ListEvents(req)
        return [Event.from_proto(x, client=self.client) for x in resp.events]

    def list_all(self, filters: Optional[list(str)] = None) -> Pager[Event]:
        """List all Events

        Pages through all events using an iterator.

        Returns:
            A iterator of all events.
        """

        def next_page(
            next_page_token,
        ):
            req = management_pb2.ListEventsRequest(
                parent=self.parent, filters=filters, page_token=next_page_token
            )
            resp = self.client._management.ListEvents(req)
            return (
                [Event.from_proto(x, client=self.client) for x in resp.events],
                resp.next_page_token,
            )

        return Pager(next_page)


class Event(Resource, types.Event):
    """Model resource."""

    # name_pattern: str = "projects/{project}/models/{model}"
    manager: EventManager

    def _init(self):
        self.manager = EventManager(parent=self.parent, client=self.client)

    def make_name(self):
        return "%s/events/%s" % (self.parent, self.id)



from __future__ import annotations
from typing import List, Optional, Iterator

from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import types
from continual.python.sdk.resource import Resource
from continual.python.sdk.manager import Manager
from continual.python.sdk.iterators import Pager
from continual.python.sdk.waiter import wait
from continual.python.sdk.events import EventManager

import requests


class BatchPredictionManager(Manager):
    """Manages ingestions resources."""

    name_pattern: str = (
        "projects/{project}/models/{model}/batchPredictions/{batch_prediction_job}"
    )

    def get(self, id: str) -> BatchPrediction:
        """Get batch prediction job.

        Arguments:
            id: Batch Prediction  name or id.

        Returns
            A Batch prediction .
        """
        req = management_pb2.BatchPredictionRequest(name=self.name(id))
        resp = self.client._management.GetBatchPrediction(req)
        return BatchPrediction.from_proto(resp, client=self.client)

    def list(
        self,
        page_size: Optional[int] = None,
        filters: List[str] = None,
        all_projects: bool = False,
    ) -> List[BatchPrediction]:
        """List batch prediction jobs.

        Arguments:
            page_size: Number of items to return.

            filter: List of filters to apply to batch prediction job. Can be:
            - state  (i.e. state:FAILED)
            - incremental (i.e. incremental:True)
            - dest type (i.e. dest_type:STORE)

        Returns:
            A list of batch prediction jobs.
        """
        req = management_pb2.ListBatchPredictionsRequest(
            parent=self.parent,
            page_size=page_size,
            filters=filters,
            all_projects=all_projects,
        )
        resp = self.client._management.ListBatchPredictions(req)
        return [
            BatchPrediction.from_proto(u, client=self.client)
            for u in resp.batch_predictions
        ]

    def list_all(self) -> Iterator[BatchPrediction]:
        """List all batch prediction jobs.

        Pages through all batch prediction jobs using an iterator.

        Returns:
            A iterator of all batch prediction jobs.
        """

        def next_page(next_page_token):
            req = management_pb2.ListBatchPredictionsRequest(
                parent=self.parent, page_token=next_page_token
            )
            resp = self.client._management.ListBatchPredictions(req)
            return (
                [
                    BatchPrediction.from_proto(u, client=self.client)
                    for u in resp.batch_predictions
                ],
                resp.next_page_token,
            )

        return Pager(next_page)

    def cancel(self, id: str) -> None:
        """Cancel an ingestion.

        Arugments:
            id: Name or id of ingestion

        """
        req = management_pb2.BatchPredictionRequest(name=self.name(id))
        self.client._management.CancelBatchPrediction(req)

    def download(self, id: str, file: str) -> None:
        """Download csv file created by batch prediction.

        Arguments:
            id: Batch Prediction  name or id.
            file: the path to a file, relative to the current working directory or absolute

        """
        batch_prediction_job = self.get(id)
        with requests.get(
            batch_prediction_job.csv_file_path, allow_redirects=True
        ) as fd:
            open(file, "xb").write(fd.content)
        return


class BatchPrediction(Resource, types.BatchPrediction):
    """BatchPrediction resource."""

    name_pattern: str = (
        "projects/{project}/models/{model}/batchPredictions/{batch_prediction_job}"
    )
    manager: BatchPredictionManager

    events: EventManager
    """Event manager."""

    def _init(self):
        self.manager = BatchPredictionManager(parent=self.parent, client=self.client)
        self.events = EventManager(parent=self.name, client=self.client)

    def cancel(self):
        """Cancel batch prediction job."""
        self.manager.cancel(self.id)

    def wait(self, echo: bool = True, timeout: int = None) -> BatchPrediction:
        """Wait for batch prediction job to complete.

        Arguments:
            echo: Display progress.
            timeout: Timeout in seconds.
        Returns:
            Updated batch prediction job.
        """
        return wait(
            getter=lambda: self.manager.get(self.id),
            echo=echo,
            timeout=timeout,
        )

    def download(self, file: str) -> None:
        """Download csv file created by batch prediction.

        Arguments:
            file: the path to a file, relative to the current working directory or absolute

        """
        self.manager.download(self.id, file)
        return



from __future__ import annotations
import grpc
from typing import Optional
from google.protobuf.empty_pb2 import Empty
from continual.python.sdk.config import Config
from continual.rpc.management.v1 import management_pb2
from continual.rpc.management.v1 import management_pb2_grpc
from continual.rpc.management.v1 import types
from continual.python.sdk.projects import ProjectManager
from continual.python.sdk.environments import EnvironmentManager
from continual.python.sdk.organizations import OrganizationManager
from continual.python.sdk.featuresets import FeatureSetManager
from continual.python.sdk.ingestions import IngestionManager
from continual.python.sdk.models import ModelManager
from continual.python.sdk.users import UserManager, User
from continual.python.sdk.model_versions import ModelVersionManager
from continual.python.sdk.experiments import ExperimentManager
from continual.python.sdk.promotions import PromotionManager
from continual.python.sdk.changes import ChangeManager
from continual.python.sdk.interceptors import AuthInterceptor
from continual.python.sdk.exceptions import normalize_exceptions_for_class
from continual.python.sdk.batchpredictions import BatchPredictionManager
from continual.python.sdk.events import EventManager


def _get_management_channel(endpoint, options=None):
    # TODO: Find a better way to configure these.
    # set max msg size to 100MB
    if options is None:
        options = [
            ("grpc.max_receive_message_length", 100 * 1024 * 1024),
            ("grpc.max_send_message_length", 100 * 1024 * 1024),
        ]
    else:
        options.extend(
            [
                ("grpc.max_receive_message_length", 100 * 1024 * 1024),
                ("grpc.max_send_message_length", 100 * 1024 * 1024),
            ]
        )

    if endpoint == "https://api.continual.ai":
        credentials = grpc.ssl_channel_credentials()
        return grpc.secure_channel("api.continual.ai:443", credentials, options)
    elif endpoint == "https://api-benchmark.continual.ai":
        credentials = grpc.ssl_channel_credentials()
        return grpc.secure_channel(
            "api-benchmark.continual.ai:443", credentials, options
        )
    elif endpoint == "http://localhost" or endpoint == "http://127.0.0.1":
        host = "127.0.0.1:30003"
        return grpc.insecure_channel(host, options)
    elif endpoint == "http://host.docker.internal":
        host = "host.docker.internal:30003"
        return grpc.insecure_channel(host, options)
    elif endpoint == "http://cluster":
        host = "management:3001"
        return grpc.insecure_channel(host, options)
    raise ValueError(
        "Unsupported private endpoint configuration.  Please contact support@continual.ai to request access to Continual Private Cloud."
    )


class Client:
    """Continual client."""

    users: UserManager
    """User manager."""

    projects: ProjectManager
    """Project manager."""

    environments: EnvironmentManager
    """Environment manager."""

    organizations: OrganizationManager
    """Organization manager."""

    feature_sets: FeatureSetManager
    """Feature set manager."""

    ingestions: IngestionManager
    """Ingestion manager"""

    models: ModelManager
    """Model manager."""

    model_versions: ModelVersionManager
    """Model versions manager."""

    experiments: ExperimentManager
    """Experiments manager."""

    promotions: PromotionManager
    """Promotion manager."""

    changes: ChangeManager
    """Change manager."""

    events: EventManager
    """Events manager."""

    batch_prediction_jobs: BatchPredictionManager
    """Batch Prediction  manager"""

    config: Config
    """Client configuration."""

    _management: management_pb2_grpc.ManagementAPIStub
    """Management GRPC API."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        project: Optional[str] = None,
        endpoint: Optional[str] = None,
    ):
        """Initialize client.

        It is recommended to use `continual login` to generate a
        local on-disk API key.

        Arguments:
            api_key: API key.
            project: Default project.
            endpoint: Cluster endpoint (Private Cloud only)
        """

        self.config = Config(api_key=api_key, project=project, endpoint=endpoint)

        # Initialize GRPC connections.
        auth_interceptor = AuthInterceptor(lambda: self.config.api_key)

        self._mgmt_channel = _get_management_channel(self.config.endpoint)

        self._mgmt_channel = grpc.intercept_channel(
            self._mgmt_channel, auth_interceptor
        )
        self._management = normalize_exceptions_for_class(
            management_pb2_grpc.ManagementAPIStub(self._mgmt_channel)
        )

        # Initialize managers.
        parent = self.config.project
        if parent is None:
            parent = "projects/-"
        else:
            env = self.config.environment
            if (
                env == None
                or env == "master"
                or env == "main"
                or env == "production"
                or len(env) == 0
            ):
                env = "prod"
            if env != "prod":
                parent = f"{self.config.project}@{self.config.environment}"

        self.users = UserManager(client=self)
        self.projects = ProjectManager(client=self)
        self.environments = EnvironmentManager(client=self, parent=parent)
        self.organizations = OrganizationManager(client=self)
        self.feature_sets = FeatureSetManager(client=self, parent=parent)
        self.models = ModelManager(client=self, parent=parent)
        self.model_versions = ModelVersionManager(
            client=self, parent=parent + "/models/-"
        )
        self.experiments = ExperimentManager(
            client=self, parent=parent + "/models/-/versions/-"
        )
        self.promotions = PromotionManager(client=self, parent=parent + "/models/-")
        self.ingestions = IngestionManager(
            client=self, parent=parent + "/featureSets/-"
        )
        self.changes = ChangeManager(client=self, parent=parent)
        self.batch_prediction_jobs = BatchPredictionManager(
            client=self, parent=parent + "/models/-"
        )
        self.events = EventManager(client=self, parent=parent)

    def __del__(self) -> None:
        """Clean up activate GRPC channels when client deleted."""
        if hasattr(self, "_mgmt_channel"):
            self._mgmt_channel.close()
        if hasattr(self, "_fs_channel"):
            self._fs_channel.close()
        if hasattr(self, "_gw_channel"):
            self._gw_channel.close()

    def viewer(self) -> User:
        """Currently authenticated user."""
        resp = self._management.GetViewer(Empty())
        return User.from_proto(resp, client=self)

    def register(
        self, first_name: str, last_name: str, email: str, password: str
    ) -> User:
        """Register a new account.

        This function registers and authenticates a new user.  This is
        meant only for automation purposes.  Please sign up directly on
        https://continual.ai instead.

        Arguments:
            first_name: First name of user.
            last_name: Last name of user
            email: Email address.
            password: Password.

        Returns:
            The newly created authenticated user.
        """
        req = management_pb2.RegisterRequest(
            first_name=first_name, last_name=last_name, email=email, password=password
        )
        resp = self._management.Register(req)
        self.config.api_key = resp.auth_token
        self.config.email = email
        return User.from_proto(resp.user, client=self)

    def login(self, email: str, password: str) -> User:
        """Login in to Continual.

        It is strongly recommended to use `continual login` CLI
        or an API key instealling logging in via the SDK.

        Args:
            email: Email address.
            password: Password.
        Returns:
            The authenticated user.
        """
        req = management_pb2.LoginRequest(email=email, password=password)
        resp = self._management.Login(req)
        self.config.api_key = resp.auth_token
        self.config.email = email
        return User.from_proto(resp.user, client=self)

    def logout(self) -> None:
        """Logout.

        Logs out current session deleting the associated API key.
        """
        if self.config.api_key is not None and self.config.api_key != "":
            self._management.Logout(Empty())
            self.config.api_key = None
            self.config.project = None
            self.config.email = None
            self.config.save()

    def infer_schema(
        self,
        project: str = None,
        table: str = None,
        query: str = None,
        file: str = None,
        type: str = "FeatureSet",
    ):
        """Infer Schema.

        Infers schema from file or data warehouse table.
        """
        if project is None:
            project = self.config.project
        elif not ("/" in project):
            project = "projects/%s" % project

        if query is not None:
            req = management_pb2.GetInferredSchemaRequest(project=project, query=query)
        elif table is not None:
            query = "SELECT * FROM %s" % table
            req = management_pb2.GetInferredSchemaRequest(project=project, query=query)
        elif file is not None:
            p = self.projects.get(project)
            (_, table) = p.seed(file)
            query = "SELECT * FROM %s" % table
            req = management_pb2.GetInferredSchemaRequest(project=project, query=query)
        else:
            return {"Error: Either 'table', 'query', or 'file' must be specified"}

        resp = self._management.GetInferredSchema(req)

        schema = []
        for c in resp.columns:
            types.ColumnConfig.from_proto(c)
            schema.append(c)

        return schema



import typer
import humanize

from continual.python.cli import utils
from continual import Client

from rich.console import Console
from datetime import datetime
from typing import List

app = typer.Typer(help="Manage promotions.")


def format_promotion_data(d, zipped=False, all_projects=False):
    end_time = d.demoted_time
    if not d.promoted_time:
        online_time = "N/A"
        end_time = "N/A"
        duration = "N/A"
    elif end_time:
        online_time = d.promoted_time.replace(microsecond=0)
        end_time = end_time.replace(microsecond=0)
        duration = humanize.naturaldelta(end_time - online_time)
    else:
        online_time = d.promoted_time.replace(microsecond=0)
        end_time = "N/A"
        if d.state.value == "SUCCEEDED":
            duration = humanize.naturaldelta(datetime.utcnow() - online_time)
        else:
            duration = "N/A"
    if zipped:
        data = [
            d.id,
            d.name,
            d.state.value,
            d.improvement_metric,
            d.improvement_metric_value,
            online_time,
            end_time,
            duration,
        ]
        headers = [
            "ID",
            "Name",
            "State",
            "Metric",
            "Metric Value",
            "Promoted",
            "Demoted",
            "Time Promoted",
        ]
        return tuple(
            [x[0], x[1]] for x in (zip(headers, data))
        )  # for some reason list(zip) causes issues, so ...
    else:
        data = [
            d.id,
            d.state.value,
            d.name.split("/")[3],
            d.model_version.split("/")[-1],
            d.improvement_metric,
            d.improvement_metric_value,
            online_time,
            end_time,
            duration,
        ]
        headers = [
            "ID",
            "State",
            "Model",
            "Model Version",
            "Metric",
            "Metric Value",
            "Promoted",
            "Demoted",
            "Time Promoted",
        ]
        if all_projects:
            data.insert(0, d.parent.split("/")[1])
            headers.insert(0, "Project")
        return (data, headers)


# use callback to run list command if nothing is passed in
@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        list(
            model=None,
            project=None,
            environment=None,
            n=30,
            filters=[],
            all_projects=False,
            style=None,
        )


@app.command("list")
@utils.exit_on_error
def list(
    model: str = typer.Option(None, help="Model ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    n: int = typer.Option(30, "--num", "-n", help="Number of records to show."),
    filters: List[str] = typer.Option([], "--filter", "-f", help="List of filters."),
    all_projects: bool = typer.Option(False, "--all", "-a", help="Show all projects."),
    style: utils.ContinualStyle = typer.Option(None, help="Color to use for list."),
):
    """List promotions.

    Filters can include:
        --state (i.e. state:FAILED)
    """
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    model_snippet = ""
    filter_snippet = " (n=%s)" % n
    project_snippet = "project %s, environment %s" % (project, environment)
    if model is not None:
        c = c.models.get(model)
        model_snippet = "for model %s " % model
    if len(filters) > 0:
        filter_snippet = " with filters %s" % str(filters) + filter_snippet
    if all_projects:
        project_snippet = "all accessible projects"
    data = []
    headers = []
    for d in c.promotions.list(n, filters=filters, all_projects=all_projects):
        (d_data, headers) = format_promotion_data(d, all_projects=all_projects)
        data.append(d_data)
    typer.secho(
        "\nFound %s promotions %sin %s%s: "
        % (len(data), model_snippet, project_snippet, filter_snippet),
        fg="blue",
    )
    utils.print_table(data, headers, style=utils.get_style(style))


@app.command("get")
@utils.exit_on_error
def get(
    promotion: str = typer.Argument(..., help="Promotion ID."),
    project: str = typer.Option(None, help="Project ID."),
    environment: str = typer.Option(None, "--env", help="Environment ID."),
    json: bool = typer.Option(False, "--json", help="Show full JSON representation."),
):
    """Get promotion details."""
    project, environment = utils.get_project_and_environment(project, environment)
    c = Client(project=utils.get_environment_name(project, environment))

    d = c.promotions.get(promotion)
    if json:
        console = Console()
        console.print(d.to_dict())
    else:
        data = format_promotion_data(d, zipped=True)
        typer.secho("\nRetrieving promotion %s: \n" % (promotion), fg="blue")
        utils.print_info(data)
